"""
The state module contains the State class and potential helper methods.
"""
from datetime import datetime
from typing import Optional


class State:
    """
    State controls the state file which is saving the last date and transaction
    hash.
    """

    filename: str
    start_date_config: Optional[datetime]

    def __init__(self, filename: str, start_date_config: Optional[datetime] = None):
        self.filename = filename
        self.start_date_config = start_date_config

    def restore(self) -> (datetime, str):
        """
        Restore returns the datetime and last hash. If no date could be found,
        it falls back to the provided fallback_date.
        """
        start_date: Optional[datetime] = None
        last_hash: Optional[str] = None
        try:
            with open(self.filename, "r") as state_file:
                contents = state_file.read().splitlines()
                start_date = datetime.fromisoformat(contents[0])
                last_hash = contents[1]
        except:  # pylint: disable=bare-except
            if self.start_date_config is not None:
                start_date = self.start_date_config
            else:
                start_date = datetime.now()

        return (start_date, last_hash)

    def store(self, last_hash: str) -> None:
        """
        Save the current date and last hash back to the file.
        """
        with open(self.filename, "w") as state_file:
            state_file.write(
                "%s\n%s" % (datetime.now().strftime("%Y-%m-%d"), last_hash,)
            )



from setuptools import setup, find_packages

from ing_ynab import VERSION

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="ing-ynab",
    version=VERSION,
    description="Import your ING Germany bank statements via FinTS into YNAB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bahlo/ing-ynab",
    author="Arne Bahlo",
    author_email="hallo@arne.me",
    license="MIT",
    keywords="ing ynab fints",
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        "bleach==3.2.1; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
        "certifi==2020.6.20",
        "chardet==3.0.4",
        "elementpath==2.0.3",
        "fints==3.0.0",
        "idna==2.10; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "mt-940==4.23.0",
        "packaging==20.4; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "pyparsing==2.4.7; python_version >= '2.6' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "python-dotenv==0.15.0",
        "requests==2.24.0",
        "sepaxml==2.3.0",
        "six==1.15.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "text-unidecode==1.3",
        "urllib3==1.25.10; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4' and python_version < '4'",
        "webencodings==0.5.1",
        "xmlschema==1.2.5",
    ],
    dependency_links=[],
    python_requires=">=3.8",
    entry_points={"console_scripts": ["ing-ynab=ing_ynab.cli:main"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Typing :: Typed",
    ],
)



"""
Provides methods to work with the YNAB API.
"""
from datetime import datetime
from decimal import Decimal
from typing import List, Optional, Dict
from mt940.models import Transaction as FinTSTransaction
import requests

YNAB_BASE_URL = "https://api.youneedabudget.com/v1"


class YNABError(Exception):
    """
    YNABError is raised when YNAB returns an error.
    """


class YNABClient:
    """
    YNABClient provides methods to transform and import transactions into YNAB.
    """

    access_token: str
    account_id: str
    budget_id: str
    flag_color: Optional[str]

    def __init__(
        self,
        access_token: str,
        account_id: str,
        budget_id: str,
        flag_color: Optional[str] = None,
    ):
        self.access_token = access_token
        self.account_id = account_id
        self.budget_id = budget_id
        self.flag_color = flag_color

    def import_transactions(self, transactions: List[Dict[str, str]]) -> List[int]:
        """
        Import the transaction into YNAB, returns an array with transaction ids.
        """
        headers = {"Authorization": "Bearer " + self.access_token}
        payload = {
            "transactions": transactions,
        }
        path = "/budgets/" + self.budget_id + "/transactions"
        response = requests.post(YNAB_BASE_URL + path, json=payload, headers=headers)
        body = response.json()
        if response.status_code >= 400:
            raise YNABError(body["error"]["detail"])
        return body["data"]["transaction_ids"]

    def transform_transactions(
        self, transactions: List[FinTSTransaction],
    ) -> List[Dict[str, str]]:
        """
        Transform the FinTS transactions into something the YNAB API understands.
        """
        transformed = []
        for transaction in transactions:
            data = transaction.data

            # ING sometimes adds transactions that are dated to the future.
            # YNAB doesn't support this, so since the transaction already
            # happened anyway, we're overriding the date.
            date = data["date"]
            if date > datetime.now():
                date = datetime.now()

            transformed.append(
                {
                    "account_id": self.account_id,
                    "date": date.isoformat(),
                    "amount": int(data["amount"].amount * Decimal("1000.0")),
                    "payee_name": data["applicant_name"],
                    "cleared": "cleared",
                    "memo": data["purpose"],
                    "flag_color": self.flag_color,
                }
            )
        return transformed



"""
Provides classes and functions to work with the ING FinTS API.
"""
import datetime
from hashlib import sha256
from typing import Optional, List
from mt940.models import Transaction
from fints.client import FinTS3PinTanClient
from fints.models import SEPAAccount

ING_DE_BLZ = "50010517"
ING_DE_ENDPOINT = "https://fints.ing-diba.de/fints/"


class AccountNotFoundException(Exception):
    """
    Raised if the account could not be found.
    """

    accounts: List[SEPAAccount] = []

    def __init__(self, accounts: List[SEPAAccount]):
        Exception.__init__()
        self.accounts = accounts


class INGClient:
    """
    INGClient provides methods to connect to ING via FinTS.
    """

    fints_client: FinTS3PinTanClient
    selected_account: SEPAAccount

    def __init__(self, login: str, pin: str, fints_product_id: Optional[str] = None):
        self.fints_client = FinTS3PinTanClient(
            ING_DE_BLZ, login, pin, ING_DE_ENDPOINT, product_id=fints_product_id,
        )

    def select_account(self, iban: str) -> None:
        """
        Select the account with the given IBAN.
        """
        accounts = self.fints_client.get_sepa_accounts()

        self.selected_account = None
        for account in accounts:
            if account.iban == iban:
                self.selected_account = account
                break

        if self.selected_account is None:
            raise AccountNotFoundException(accounts)

    def get_transactions(
        self,
        start_date: Optional[datetime.datetime] = None,
        last_hash: Optional[str] = None,
    ) -> List[Transaction]:
        """
        Get transactions for the selected account.
        """
        transactions = self.fints_client.get_transactions(
            self.selected_account, start_date=start_date
        )

        array_start = 0
        for i, transaction in enumerate(transactions):
            if hash_transaction(transaction) == last_hash:
                array_start = i + 1
                break

        transactions = transactions[array_start:]
        return transactions


def hash_transaction(transaction: Transaction) -> str:
    """
    Generate a hash for the transaction to make them identifiable.
    """
    data = transaction.data
    payload = "%s:%s:%s:%s:%s" % (
        data["date"],
        data["applicant_name"],
        data["purpose"],
        data["amount"],
        data.get("end_to_end_reference"),
    )
    return sha256(payload.encode("utf-8")).hexdigest()



"""
This module contains the core business logic.
"""
from datetime import datetime
from time import sleep
from typing import NoReturn
import logging
import os
import sys
from getpass import getpass
from dotenv import load_dotenv

from ing_ynab.ynab import YNABClient, YNABError
from ing_ynab.ing import INGClient, AccountNotFoundException, hash_transaction
from ing_ynab.state import State


def ing_to_ynab(
    state: State, ing_client: INGClient, ynab_client: YNABClient, debug: bool = False,
) -> NoReturn:
    """
    This code is called in a predefined interval to add new ing transactions
    into ynab.
    """
    start_date, last_hash = state.restore()

    transactions = ing_client.get_transactions(
        start_date=start_date, last_hash=last_hash
    )

    if len(transactions) == 0:
        print("No new transactions found")
        return

    # Transform FinTS transactions to YNAB transactions.
    ynab_transactions = ynab_client.transform_transactions(transactions)

    # Print or import the transformed transactions.
    if debug:
        for transaction in ynab_transactions:
            print(transaction)
    else:
        imported = ynab_client.import_transactions(ynab_transactions)
        print("Imported %d new transaction(s)" % len(imported))

    state.store(hash_transaction(transactions[len(transactions) - 1]))


def main() -> int:
    """
    Start the main business logic.
    """
    load_dotenv()

    # Parse environment variables
    debug = os.environ.get("DEBUG") == "1"
    if debug:
        logging.basicConfig(level=logging.DEBUG)

    ing_pin = os.environ.get("ING_PIN")
    if ing_pin is None:
        ing_pin = getpass("ING pin: ")

    ynab_access_token = os.environ.get("YNAB_ACCESS_TOKEN")
    if ynab_access_token is None and not debug:
        ynab_access_token = getpass("YNAB Access Token: ")

    ynab_account_id = os.environ["YNAB_ACCOUNT_ID"]
    ynab_budget_id = os.environ["YNAB_BUDGET_ID"]
    ynab_flag_color = os.environ.get("YNAB_FLAG_COLOR")

    start_date = None
    start_date_env = os.environ.get("START_DATE")
    if start_date_env is not None:
        start_date = datetime.fromisoformat(start_date_env)

    interval = int(os.environ.get("SLEEP_INTERVAL", "300"))

    # Create INGClient
    ing_client = INGClient(
        os.environ["ING_LOGIN"],
        ing_pin,
        fints_product_id=os.environ.get("FINTS_PRODUCT_ID"),
    )

    # Select account with configured IBAN
    try:
        ing_client.select_account(os.environ["ING_IBAN"])
    except AccountNotFoundException as ex:
        print("Could not find account, is the IBAN correct?")
        print("Available accounts: %s" % ex.accounts)
        return 1

    ynab_client = YNABClient(
        ynab_access_token, ynab_account_id, ynab_budget_id, flag_color=ynab_flag_color
    )

    state = State("state", start_date)

    # Import new statements into YNAB every n minutes
    while True:
        try:
            ing_to_ynab(state, ing_client, ynab_client, debug=debug)
        except YNABError as ex:
            print("Could not import transactions: %s" % ex)
        except KeyboardInterrupt:
            raise  # We need to have this case for ^C to work
        except:  # pylint: disable=bare-except
            print("Unexpected error:", sys.exc_info()[0])
        print("Sleeping for %d seconds" % interval)
        sleep(interval)

    return 0
